# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# Copyright (c) 2015, Arista Networks, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
#   Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#
#   Neither the name of Arista Networks nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ARISTA NETWORKS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import shutil
import sys

from glob import glob
from ztpserver import config

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

def install():
    if "install" in sys.argv:
        return True
    else:
        return False

def join_url(x, y):
    return '/' + '/'.join([z for z in x.split('/') + y.split('/') if z])

def ensure_dir(f):
    d = os.path.dirname(f)
    if not os.path.exists(d):
        os.makedirs(d)

conf_path = config.CONF_PATH
install_path = config.INSTALL_PATH

if install() and os.environ.get('ZTPS_INSTALL_PREFIX'):
    print "Customizing install for VirtualEnv install with RPM"
    conf_path = join_url(os.environ.get('ZTPS_INSTALL_PREFIX'), config.CONF_PATH)
    install_path = join_url(os.environ.get('ZTPS_INSTALL_PREFIX'), config.INSTALL_PATH)

packages = ['ztpserver']
if install() and os.environ.get('READTHEDOCS'):
    print 'Customizing install for ReadTheDocs.org build servers...'
    conf_path = '.' + conf_path
    install_path = '.' +  install_path
    os.environ['ZTPS_INSTALL_ROOT'] = '.'
    from subprocess import call
    call(['docs/setup_rtd_files.sh'])
    packages.append('client')
    packages.append('actions')
    packages.append('plugins')

install_requirements = None
version = None

install_requirements = open('requirements.txt').read().split('\n')
install_requirements = [x.strip() for x in install_requirements
                        if x.strip() and
                        'dev only' not in x]
version = open('VERSION').read().split()[0].strip()

data_files = []
# configuration folders are not cleared on upgrade/downgrade
for folder in ['nodes', 'definitions', 'files', 'resources',
               'bootstrap', 'config-handlers']:
    path = '%s/%s' % (install_path, folder)
    if install() and not os.path.isdir(path):
        if os.path.exists(path):
            os.remove(path)
        data_files += [(path, [])]

for (filename, dst, src) in [('neighbordb',
                              install_path,
                              'conf/neighbordb'),
                             ('bootstrap.conf',
                              '%s/bootstrap' % install_path,
                              'conf/bootstrap.conf'),
                             ('ztpserver.conf',
                              conf_path,
                              'conf/ztpserver.conf'),
                             ('ztpserver.wsgi',
                              conf_path,
                              'conf/ztpserver.wsgi')]:
    file_path = '%s/%s' % (dst, filename)
    if install() and os.path.exists(file_path):
        if os.path.isdir(file_path):
            shutil.rmtree(file_path,
                          ignore_errors=True)
        else:
            # do this manually
            shutil.copy(src, file_path + '.new')
            continue

    data_files += [(dst, glob(src))]

# bootstrap file, libraries, VERSION, plugins and actions are
# always overwritten
file_list = [('bootstrap', '%s/bootstrap' % install_path,
              'client/bootstrap')]
for filename in glob('actions/*'):
    file_list += [(filename.split('/')[-1],
                   '%s/actions' % install_path,
                   filename)]
for filename in glob('plugins/*'):
    file_list += [(filename.split('/')[-1],
                   '%s/plugins' % install_path,
                   filename)]
for filename in glob('client/lib/*'):
    file_list += [(filename.split('/')[-1],
                   '%s/files/lib' % install_path,
                   filename)]
for (filename, dst, src) in file_list:
    file_path = '%s/%s' % (dst, filename)
    if install() and os.path.exists(file_path) and \
            os.path.isdir(file_path):
        shutil.rmtree(file_path,
                      ignore_errors=True)
    data_files += [(dst, glob(src))]

setup(
    name='ztpserver',
    version=version,
    description = 'ZTP Server for EOS',
    author='Arista Networks',
    author_email='eosplus-dev@arista.com',
    url='https://github.com/arista-eosplus/ztpserver',
    download_url='https://github.com/arista-eosplus/ztpserver/tarball/v%s' % \
                  version,
    license='BSD-3',
    install_requires=install_requirements,
    packages=packages,
    scripts=glob('bin/*'),
    data_files=data_files
)

# hidden version file
if install():
    custom_path = os.environ.get('ZTPS_INSTALL_ROOT')
    if custom_path:
        version_file =  join_url(custom_path, config.VERSION_FILE_PATH)[1:]
    else:
        version_file =  config.VERSION_FILE_PATH
    ensure_dir(version_file)
    shutil.copy('VERSION', version_file)



#
# Copyright (c) 2014, Arista Networks, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
#   Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#
#   Neither the name of Arista Networks nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ARISTA NETWORKS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# pylint: disable=W0631

import string    #pylint: disable=W0402
import re
import inspect
import logging
import collections

from ztpserver.utils import expand_range, parse_interface
from ztpserver.config import runtime

REQUIRED_PATTERN_ATTRIBUTES = ['name', 'definition']
OPTIONAL_PATTERN_ATTRIBUTES = ['node', 'variables', 'interfaces']
INTERFACE_PATTERN_KEYWORDS = ['any', 'none']
ANTINODE_PATTERN = r'[^%s]' % string.hexdigits
KW_ANY_RE = re.compile(r' *any *')
KW_NONE_RE = re.compile(r' *none *')
WC_PORT_RE = re.compile(r'.*')

INVALID_INTERFACE_PATTERNS = [(KW_ANY_RE, KW_ANY_RE, KW_NONE_RE),
                              (KW_ANY_RE, KW_NONE_RE, KW_NONE_RE),
                              (KW_ANY_RE, KW_NONE_RE, KW_ANY_RE),
                              (KW_ANY_RE, KW_NONE_RE, WC_PORT_RE),
                              (KW_NONE_RE, KW_ANY_RE, KW_ANY_RE),
                              (KW_NONE_RE, KW_ANY_RE, KW_NONE_RE),
                              (KW_NONE_RE, KW_NONE_RE, WC_PORT_RE),
                              (KW_NONE_RE, KW_NONE_RE, KW_ANY_RE)]


log = logging.getLogger(__name__)   #pylint: disable=C0103


class ValidationError(Exception):
    ''' Base error class for validation failures '''
    pass


class Validator(object):

    def __init__(self, node_id):
        self.node_id = node_id
        self.data = dict()
        self.fail = False
        self.errors = list()

    def validate(self, data=None):
        log.debug('%s: running %s.validate' % 
                  (self.node_id, self.__class__.__name__))
        if data:
            self.data = data
        else:
            self.data = dict()

        error = None
        methods = inspect.getmembers(self, predicate=inspect.ismethod)
        for name in methods:
            if name[0].startswith('validate_'):
                if 'name' not in self.data:
                    name_string  = ''
                else:
                    name_string  = 'for \'%s\'' % self.data['name']

                log.debug('%s: running %s.%s %s' % 
                          (self.node_id, self.__class__.__name__,
                           name[0], name_string))
                try:
                    getattr(self, name[0])()
                except ValidationError as err:
                    if not error:
                        error = err
        if error:
            self.error(err)

        return not self.fail

    def error(self, err, *args, **kwargs):
        #pylint: disable=W0613
        cls = str(self.__class__).split('\'')[1].split('.')[-1]
        log.error('%s: %s validation error: %s' % 
                  (self.node_id, cls, err))
        self.fail = True


class NeighbordbValidator(Validator):

    def __init__(self, node_id):
        self.invalid_patterns = set()
        self.valid_patterns = set()
        super(NeighbordbValidator, self).__init__(node_id)

    def validate_variables(self):
        variables = self.data.get('variables', None)
        if variables is not None:
            if not hasattr(variables, '__iter__'):
                raise ValidationError('invalid global variables value (%s)' %
                                      variables)

    def validate_patterns(self):
        patterns = self.data.get('patterns', None)

        if not patterns:
            log.warning('%s: no patterns found in neighbordb (%s)' % 
                        (self.node_id, self.data))
            return

        for index, entry in enumerate(patterns):
            name = entry.get('name', None)
 
            validator = PatternValidator(self.node_id)

            if name and validator.validate(entry):
                log.debug('%s: adding pattern \'%s\' (%s) to valid patterns' % 
                         (self.node_id, name, entry))
                self.valid_patterns.add((index, str(name)))
            else:
                if not name:
                    name = 'N/A'
                log.debug('%s: adding pattern \'%s\' (%s) to '
                          'invalid patterns' % 
                         (self.node_id, name, entry))
                self.invalid_patterns.add((index, str(name)))
        
        if self.invalid_patterns:
            raise ValidationError('invalid patterns: %s' % 
                                  self.invalid_patterns)


class PatternValidator(Validator):

    def __init__(self, node_id):
        self.invalid_interface_patterns = set()
        self.valid_interface_patterns = set()
        super(PatternValidator, self).__init__(node_id)

    def validate_attributes(self):
        for attr in REQUIRED_PATTERN_ATTRIBUTES:
            if attr not in self.data:
                raise ValidationError('missing attribute: %s' % attr)

        if 'node' not in self.data and 'interfaces' not in self.data:
            raise ValidationError('missing attribute: \'node\' OR '
                                  '\'interfaces\'')

        for attr in OPTIONAL_PATTERN_ATTRIBUTES:
            if attr not in self.data:
                log.warning('%s: PatternValidator warning: \'%s\' is missing '
                            'optional attribute (%s)' % 
                            (self.node_id, self.data['name'], attr))

        

    def validate_name(self):
        if not self.data or 'name' not in self.data:
            raise ValidationError('missing attribute: \'name\'')

        if  self.data['name'] is None or not isinstance(self.data['name'], 
                                                        (int, basestring)):
            raise ValidationError('invalid value for \'name\' (%s)' %
                                  self.data['name'])

    def validate_interfaces(self):
        if not self.data:
            return

        if 'interfaces' not in self.data:
            return

        if not isinstance(self.data['interfaces'], 
                          collections.Iterable):
            raise ValidationError('\'interfaces\' is not iterable (%s)' %
                                  self.data['interfaces'])

        for index, pattern in enumerate(self.data['interfaces']):
            if not isinstance(pattern, collections.Mapping):
                raise ValidationError('invalid value for interface pattern '
                                      '(%s)' % pattern)

            validator = InterfacePatternValidator(self.node_id)

            if validator.validate(pattern):
                log.debug('%s: adding interface pattern \'%s\' to '
                         'valid interface patterns' % 
                         (self.node_id, repr(pattern)))
                self.valid_interface_patterns.add((index, repr(pattern)))
            else:
                log.debug('%s: adding interface pattern \'%s\' to '
                         'invalid interface patterns' % 
                         (self.node_id, repr(pattern)))
                self.invalid_interface_patterns.add((index, repr(pattern)))

        if self.invalid_interface_patterns:
            raise ValidationError('invalid interface patterns: %s' %
                                  self.invalid_interface_patterns)

    def validate_definition(self):
        if not self.data:
            return

        if 'definition' not in self.data:
            return

        if not isinstance(self.data['definition'], (int, basestring)):
            raise ValidationError('invalid value for \'definition\' (%s)' %
                                  self.data['definition'])

        for wspc in string.whitespace:
            if wspc in self.data['definition']:
                raise ValidationError('invalid value for \'definition\' (%s) - '
                                      '\'%s\' not allowed' % 
                                      (self.data['definition'], wspc))

    def validate_node(self):
        if not self.data:
            return

        node = self.data.get('node', None)
        if not node:
            return
        else:
            if isinstance(node, int):
                node = str(node)

            if not isinstance(node, str):
                raise ValidationError('invalid value for \'node\' (%s)' %
                                      str(node))

        # if system MAC is used
        if runtime.default.identifier == 'systemmac':
            node = node.replace(':', '').replace('.', '')
            if re.search(ANTINODE_PATTERN, node):
                raise ValidationError('invalid value for \'node\' (%s)' %
                                      node)

    def validate_variables(self):
        if not self.data:
            return

        if 'variables' not in self.data:
            return

        if 'variables' in self.data:
            if not hasattr(self.data['variables'], '__iter__'):
                raise ValidationError('invalid value for \'variables\' ('
                                      'expecting iterable object, got: %s)' % 
                                      self.data['variables'])


class InterfacePatternValidator(Validator):

    def __init__(self, node_id):
        super(InterfacePatternValidator, self).__init__(node_id)

    def validate_interface_pattern(self):

        for interface, peer in self.data.items():
            if peer is None:
                raise ValidationError('missing peer for interface %s' % 
                                      interface)

            try:
                (device, port) = parse_interface(peer, self.node_id)
            except Exception as err:
                raise ValidationError('PatternError: %s' % err)

            if interface not in INTERFACE_PATTERN_KEYWORDS:
                try:
                    for entry in expand_range(interface):
                        self._validate_pattern(entry, device, port)
                except Exception as err:
                    raise ValidationError('invalid interface %s (%s)' % 
                                          (interface, err))
            else:
                self._validate_pattern(interface, device, port)

    def _validate_pattern(self, interface, device, port):
        # pylint: disable=R0201
        if KW_NONE_RE.match(interface) and KW_NONE_RE.match(device) \
                and KW_NONE_RE.match(port):
            # no LLDP neighbors
            return

        for interface_re, device_re, port_re in INVALID_INTERFACE_PATTERNS:
            if interface_re.match(interface) and device_re.match(device) \
               and port_re.match(port):
                raise ValidationError('invalid interface pattern: (%s, %s, %s) '
                                      'matches (%s, %s, %s)'%
                                      (interface, device, port,
                                       interface_re.pattern, 
                                       device_re.pattern, 
                                       port_re.pattern))


def _validator(contents, cls, node_id):
    try:
        validator = cls(node_id)
        result = validator.validate(contents)
        if result:
            log.debug('%s: %s validation successful' % 
                      (node_id, validator.__class__.__name__))
        else:
            log.debug('%s: %s validation failed' % 
                      (node_id, validator.__class__.__name__))

        return result
    except Exception as exc:
        log.error('%s: failed to run validator %s(%s): %s' %
                  (node_id, cls.__name__, contents, repr(exc)))
        raise

def validate_neighbordb(contents, node_id):
    return _validator(contents, NeighbordbValidator, node_id)

def validate_pattern(contents, node_id):
    return _validator(contents, PatternValidator, node_id)



#
# Copyright (c) 2015, Arista Networks, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
#   Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#
#   Neither the name of Arista Networks nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ARISTA NETWORKS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
# pylint: disable=W0142
#
import unittest

from mock import Mock

from ztpserver.app import enable_handler_console    # pylint: disable=W0611
from ztpserver.topology import InterfacePattern, InterfacePatternError
from ztpserver.topology import Pattern, PatternError
from ztpserver.topology import Node, NodeError, Neighbor

from ztpserver.topology import ExcludesFunction, IncludesFunction
from ztpserver.topology import ExactFunction, RegexFunction

from server_test_lib import random_string, enable_logging
from server_test_lib import create_node


class NodeUnitTests(unittest.TestCase):

    def test_create_node_success(self):
        node = create_node()
        systemmac = node.systemmac
        kwargs = node.as_dict()
        node = Node(**kwargs)
        self.assertEqual(node.systemmac, systemmac)

    def test_create_node_systemmac_only(self):
        systemmac = random_string()
        node = Node(systemmac=systemmac)
        self.assertEqual(node.systemmac, systemmac)

    @classmethod
    def test_create_node_failure(cls):
        try:
            _ = Node()
        except TypeError:
            pass

    def test_create_node_neighbors_valid(self):
        nodeattrs = create_node()

        remote_device = random_string()
        remote_interface = random_string()

        neighbors = {'Ethernet1': {'device': remote_device, 
                                   'port': remote_interface}}
        nodeattrs.add_neighbors(neighbors)

        kwargs = nodeattrs.as_dict()
        node = Node(**kwargs)

        self.assertIsNotNone(node.neighbors('Ethernet1'))
        self.assertEqual(node.neighbors('Ethernet1')[0].device, 
                         remote_device)
        self.assertEqual(node.neighbors('Ethernet1')[0].interface, 
                         remote_interface)

    def test_create_node_neighbors_remote_interface_missing(self):
        nodeattrs = create_node()

        remote_device = random_string()

        neighbors = {'Ethernet1': {'remote_device': remote_device}}
        nodeattrs.add_neighbors(neighbors)

        kwargs = nodeattrs.as_dict()
        try:
            node = None
            node = Node(**kwargs)
        except NodeError:
            pass
        finally:
            self.assertIsNone(node)

    def test_create_node_neighbors_remote_device_missing(self):
        nodeattrs = create_node()

        remote_interface = random_string()

        neighbors = {'Ethernet1': {'remote_interface': remote_interface}}
        nodeattrs.add_neighbors(neighbors)

        kwargs = nodeattrs.as_dict()
        try:
            node = None
            node = Node(**kwargs)
        except NodeError:
            pass
        finally:
            self.assertIsNone(node)

    def test_add_neighbor(self):
        systemmac = random_string()
        peer = Mock()
        intf = random_string()

        node = Node(systemmac=systemmac)
        node.add_neighbor(intf, [dict(device=peer.remote_device, 
                                      port=peer.remote_interface)])

        self.assertIsNotNone(node.neighbors(intf))

        self.assertEqual(node.neighbors(intf)[0].device, 
                         peer.remote_device)
        self.assertEqual(node.neighbors(intf)[0].interface, 
                         peer.remote_interface)

    def test_add_neighbor_existing_interface(self):
        systemmac = random_string()
        peer = Mock()
        intf = random_string()

        node = Node(systemmac=systemmac)
        node.add_neighbor(intf, [dict(device=peer.remote_device, 
                                      port=peer.remote_interface)])
        self.assertRaises(NodeError, node.add_neighbor,
                          intf, [dict(device=peer.remote_device, 
                                      port=peer.remote_interface)])

    def test_add_neighbors_success(self):
        nodeattrs = create_node()

        remote_device = random_string()
        remote_interface = random_string()
        neighbors = {'Ethernet1': [{'device': remote_device, 
                                    'port': remote_interface}]}

        kwargs = nodeattrs.as_dict()
        node = Node(**kwargs)
        node.add_neighbors(neighbors)

        self.assertIsNotNone(node.neighbors('Ethernet1'))
        self.assertEqual(node.neighbors('Ethernet1')[0].device, 
                         remote_device)
        self.assertEqual(node.neighbors('Ethernet1')[0].interface, 
                         remote_interface)

    def test_serialize_success(self):
        nodeattrs = create_node()
        kwargs = nodeattrs.as_dict()
        node = Node(**kwargs)
        result = node.serialize()
        self.assertEqual(result, nodeattrs.as_dict())


class FunctionsUnitTests(unittest.TestCase):

    def test_exactfunction_true(self):
        value = random_string()
        func = ExactFunction(value)
        self.assertTrue(func.match(value))

    def test_exactfunction_false(self):
        value = random_string()
        func = ExactFunction(value)
        self.assertFalse(func.match(random_string()))

    def test_includesfunction_true(self):
        value = random_string()
        func = IncludesFunction(value)
        self.assertTrue(func.match(value))

    def test_includesfunction_false(self):
        value = random_string()
        func = IncludesFunction(value)
        self.assertFalse(func.match(random_string()))

    def test_excludesfunction_true(self):
        value = random_string()
        func = ExcludesFunction(value)
        self.assertTrue(func.match(random_string()))

    def test_excludesfunction_false(self):
        value = random_string()
        func = ExcludesFunction(value)
        self.assertFalse(func.match(value))

    def test_regexfunction_true(self):
        value = r'[\w+]'
        func = RegexFunction(value)
        self.assertTrue(func.match(random_string()))

    def test_regexfunction_false(self):
        value = '[^a-zA-Z0-9]'
        func = RegexFunction(value)
        self.assertFalse(func.match(random_string()))


class TestPattern(unittest.TestCase):

    def test_create_pattern_with_defaults(self):
        obj = Pattern(None, None, None)
        self.assertIsInstance(obj, Pattern)

    def test_create_pattern_with_kwargs(self):
        obj = Pattern(name='test',
                                         definition='test',
                                         node='abc123',
                                         variables={'var': 'test'},
                                         interfaces=[{'Ethernet1': 'any'}])
        self.assertEqual(obj.name, 'test')
        self.assertEqual(obj.definition, 'test')
        self.assertEqual(obj.node, 'abc123')
        self.assertDictEqual({'var': 'test'}, obj.variables)
        self.assertEqual(1, len(obj.interfaces))

    def test_add_interface(self):
        obj = Pattern(interfaces=[{'Ethernet1': 'any'}])
        self.assertEqual(len(obj.interfaces), 1)


class PatternUnitTests(unittest.TestCase):

    def test_create_pattern(self):
        pattern = Pattern(random_string())
        self.assertIsInstance(pattern, Pattern)

    def test_create_pattern_kwargs(self):
        kwargs = dict(name=random_string(),
                      definition=random_string(),
                      interfaces=None)

        pattern = Pattern(**kwargs)
        self.assertIsInstance(pattern, Pattern)

    @classmethod
    def test_add_interface_success(cls):
        kwargs = dict(name=random_string(),
                      definition=random_string(),
                      interfaces=None)

        pattern = Pattern(**kwargs)

        remote_remote_device = random_string()
        remote_intf = random_string()
        neighbors = dict(Ethernet1={'device': remote_remote_device,
                                    'port': remote_intf})

        pattern.add_interface(neighbors)
            
    def test_add_interface_failure(self):
        kwargs = dict(name=random_string(),
                      definition=random_string(),
                      interfaces=None)

        pattern = Pattern(**kwargs)
        self.assertRaises(PatternError, pattern.add_interface, 
                          random_string())


class TestInterfacePattern(unittest.TestCase):

    def test_create_interface_pattern(self):
        intf = 'Ethernet1'
        remote_device = random_string()
        remote_interface = random_string()

        obj = InterfacePattern(intf, remote_device,
                               remote_interface,
                               random_string())
        reprobj = 'InterfacePattern(interface=%s, remote_device=%s, ' \
                   'remote_interface=%s)' % \
                  (intf, remote_device, remote_interface)
        self.assertEqual(repr(obj), reprobj)

    def test_match_success(self):
        interface = random_string()
        remote_device = random_string()
        remote_interface = random_string()

        neighbor = Neighbor(remote_device, remote_interface)
        for intf in ['any', interface]:
            for remote_d in ['any', remote_device]:
                for remote_i in ['any', remote_interface]:
                    
                    pattern = InterfacePattern(intf, remote_d, remote_i,
                                               random_string())
                    result = pattern.match(interface, [neighbor])
                    self.assertTrue(result)

    def test_match_failure(self):
        interface = random_string()
        remote_device = random_string()
        remote_interface = random_string()

        for intf in ['none', interface + 'dummy']:
            pattern = InterfacePattern(intf, remote_device, 
                                       remote_interface,
                                       random_string())
            neighbor = Neighbor(remote_device, remote_interface)
            result = pattern.match(interface, [neighbor])
            self.assertFalse(result)

        for remote_d in ['none', remote_device + 'dummy']:
            pattern = InterfacePattern(interface, remote_d, 
                                       remote_interface,
                                       random_string())
            neighbor = Neighbor(remote_device, remote_interface)
            result = pattern.match(interface, [neighbor])
            self.assertFalse(result)

        for remote_i in ['none', remote_interface + 'dummy']:
            pattern = InterfacePattern(interface, remote_device, 
                                       remote_i, random_string())
            neighbor = Neighbor(remote_device, remote_interface)
            result = pattern.match(interface, [neighbor])
            self.assertFalse(result)

        for remote_d in ['none', remote_device + 'dummy']:
            for remote_i in ['none', remote_interface + 'dummy']:
                pattern = InterfacePattern(interface, remote_d, 
                                           remote_i, random_string())
                neighbor = Neighbor(remote_device, remote_interface)
                result = pattern.match(interface, [neighbor])
                self.assertFalse(result)

        for intf in ['none', interface + 'dummy']:
            for remote_i in ['none', remote_interface + 'dummy']:
                pattern = InterfacePattern(intf, remote_device, 
                                           remote_i, random_string())
                neighbor = Neighbor(remote_device, remote_interface)
                result = pattern.match(interface, [neighbor])
                self.assertFalse(result)

        for intf in ['none', interface + 'dummy']:
            for remote_d in ['none', remote_device + 'dummy']:
                pattern = InterfacePattern(intf, remote_d, 
                                           remote_interface, random_string())
                neighbor = Neighbor(remote_device, remote_interface)
                result = pattern.match(interface, [neighbor])
                self.assertFalse(result)

        for intf in ['none', interface + 'dummy']:
            for remote_d in ['none', remote_device + 'dummy']:
                for remote_i in ['none', remote_interface + 'dummy']:
                    pattern = InterfacePattern(intf, remote_d, 
                                               remote_i, random_string())
                    neighbor = Neighbor(remote_device, remote_interface)
                    result = pattern.match(interface, [neighbor])
                    self.assertFalse(result)

    def compile_known_function(self, interface, cls):
        pattern = InterfacePattern(random_string(),
                                   interface,
                                   random_string(),
                                   random_string())
        self.assertIsInstance(pattern.remote_device_re, cls)

    def test_compile_exact_function(self):
        interface = 'exact(\'%s\')' % random_string()
        self.compile_known_function(interface, ExactFunction)

    def test_compile_includes_function(self):
        interface = 'includes(\'%s\')' % random_string()
        self.compile_known_function(interface, IncludesFunction)

    def test_compile_excludes_function(self):
        interface = 'excludes(\'%s\')' % random_string()
        self.compile_known_function(interface, ExcludesFunction)

    def test_compile_regex_function(self):
        interface = 'regex(\'%s\')' % random_string()
        self.compile_known_function(interface, RegexFunction)

    def test_compile_no_function(self):
        interface = random_string()
        self.compile_known_function(interface, ExactFunction)

    def test_compile_unknown_function(self):
        interface = '%s(\'%s\')' % (random_string(), random_string())
        self.assertRaises(InterfacePatternError,
                          InterfacePattern,
                          random_string(), interface,
                          random_string(), random_string())

if __name__ == '__main__':
    enable_logging()
    unittest.main()
